//
// Copyright (c) 2015-2020 CNRS INRIA

#ifndef __pinocchio_spatial_log_hpp__
#define __pinocchio_spatial_log_hpp__

namespace pinocchio
{

  template<typename Scalar> struct log3_impl;
  template<typename Scalar> struct Jlog3_impl;

  template<typename Scalar> struct log6_impl;
  template<typename Scalar> struct Jlog6_impl;

} // namespace pinocchio

#endif // ifndef __pinocchio_spatial_log_hpp__
